/* -------------------------------------------------------------

 */

#include <common.h>
#include <command.h>
#include <asm/io.h>

#define POLYNOMIAL 0xA097

static unsigned short CrcTable[256]; /* CRC lookup table */



/* VPD vectors */
#define VPD_FILLER                    0x00
#define VPD_CARD_ID                   0x01
#define VPD_REVISION_ID               0x02
#define VPD_CRC                       0x03
#define VPD_CARD_BASE_MAC             0xA0
#define VPD_NUMBER_MAC_ADDRESSES      0xA1
#define VPD_CPU_ID                    0xA2
#define VPD_CPU_TYPE                  0xA3
#define VPD_INT_LEVEL                 0xB0
#define VPD_SLOTS_SUPPORTED           0xC0
#define VPD_SN                        0xC1
#define VPD_TYPE                      0xC2
#define VPD_MODEL                     0xC3
#define VPD_MFGR                      0xC4
#define VPD_PN                        0xC5
#define VPD_FRU                       0xC6
#define VPD_DESC                      0xC7
#define VPD_MAINT_LVL                 0xC8
#define VPD_OID                       0xC9
#define VPD_PASSWORD                  0xD0
#define VPD_SERVICE_TAG               0xD1
#define VPD_HW_REV                    0xD2
#define VPD_BOARD_ID                  0xD3
#define VPD_OPERATIONAL_MODE_LOCK     0xE1
#define VPD_EXTENDED_VECTOR_CODE      0xFF

/* Pack the EEPROM */

typedef struct vpd_mapping_t
{
  unsigned char length;
  unsigned char vector;
  unsigned char data[4];
} vpd_mapping;






/************************************************************************
** NAME: UpdateCRC
**
** This function updates existing CRC
**
** INPUT:
**	DataByte - Received data byte
**
** INPUT/OUTPUT:
**	crc - Current value of CRC.
************************************************************************/
static void UpdateCRC (unsigned char DataByte, unsigned short *crc)
{
  unsigned long new_crc;

  new_crc = *crc;
  new_crc = (new_crc << 8) ^ (CrcTable[new_crc >> 8] ^ DataByte);
  *crc = (unsigned short) new_crc;
}


/************************************************************************
** NAME: GenCrcTable
**
** This function computes CRC look-up table.
************************************************************************/
static void GenCrcTable (void)
{
  unsigned short CrcCode;
  unsigned char  DataByte;
  unsigned long  index, BitNo;
  unsigned long  CarryBit, DataBit;

  for( index = 0; index <= 255; index++ )
  {
    CrcCode = 0;
    DataByte = index;

    for( BitNo = 1; BitNo <= 8; BitNo++ )
    {
      CarryBit = ((CrcCode & 0x8000) != 0);
      DataBit = ((DataByte & 0x80) != 0);
      CrcCode = CrcCode << 1;
      if( CarryBit ^ DataBit )
        CrcCode = CrcCode ^ POLYNOMIAL;
      DataByte = DataByte << 1;
    }

    CrcTable [index] = CrcCode;
  }
}




/**************************************************************************
*
* @purpose  Calculate the CRC of a given null terminated string of data
*
* @param    fp pointer of buffer to calculate CRC
*
* @returns  calculated CRC
*
* @end
*
*************************************************************************/
static unsigned short calc_crc(char *fp, int len)
{
  unsigned short crc;
  unsigned int   i;

  GenCrcTable ();
  crc = 0;

  for( i=0; i<len; i++ )
  {
    UpdateCRC (fp[i], &crc);
  }

  return(crc);
}


int vpdRead(char *from, char *to)
{
  unsigned int i;
  unsigned short crc, expected_crc;


  /* read the entire vpd into memory and check the CRC */
  for( i=0; i<VPD_SIZE;i++ )
    to[i] = from[i];
#if 0
  printf(" vpdRead: %d bytes read from x%08lx to x%08lx \n", i, from, to);
#endif

#if 1 /* for now, ignore crc checking until in sync with diags */
  /* verify the CRC */
  crc = calc_crc(&to[2], 254);
  expected_crc  = ((unsigned short)(to[0])) << 8;
  expected_crc += to[1];
#if 1
  printf(" vpdRead: calc crc on 254 bytes = 0x%04x, actual crc = x%04x\n", crc, expected_crc);
#endif
  if( crc != expected_crc )
    return -1;
#endif

  return 0;
}


int getVPDField(unsigned char field, char *pVPD, void *pbuffer)
{
  vpd_mapping *p;

  /* pVPD = beginning of VPD  */
  /* field = what the user is looking for        */
  /* pbuffer = where to put data if found in vpd */

  p = (vpd_mapping *)( pVPD + 2 ); /* move past the CRC bytes */

  while( p->vector != 0x00 )
  {
    if( p->vector == field )
    {
      switch( field )
      {
        case VPD_CARD_BASE_MAC:
          memcpy( pbuffer, (void *)&p->data, 6);
          return(0);

        case VPD_SN:
          strcpy( (char *)pbuffer, (char *)&p->data );
          return(0);

        case VPD_SERVICE_TAG:
          strcpy( (char *)pbuffer, (char *)&p->data);
          return(0);

        default:
          break;
      }
    }
    /* move to next field */
    p = ( vpd_mapping * )( ( char * )p + p->length );

    /*
    * Even though a corrupted NVRAM SRAM may pass the CRC check,
    * protect ourselves from NVRAM SRAM that have garbage in them.
    */

    if( ( char * )p > ( pVPD + VPD_SIZE ) )
    {
      break;
    }
  }
  return(-1);

}




int vpdMacAddrGet(char* vpd, char *macString)
{
  /* Read box mac address */
  return getVPDField(VPD_CARD_BASE_MAC, vpd, macString);
}

